import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

class App {
    constructor() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        //renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.setPixelRatio(window.devicePixelRatio);
        //renderer.shadowMap.enabled =  true;
        domWebGL.appendChild(renderer.domElement);

        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0xf0f0f0);
        this._scene = scene;

        this._setupCamera();
        this._setupLight();
        this._setupInteractions();
        this._setupModel();
        this._setupGUI();
        //this._setupControls();
        //this._setupHelper();

        window.onresize = this.resize.bind(this);

        this.resize();
        this.render()
    }

    _setupInteractions() {
        this._raycaster = new THREE.Raycaster();
        this._pointer = new THREE.Vector2();

        this._objects = [];

        document.addEventListener("pointermove", this._onPointerMove.bind(this));
        document.addEventListener("pointerdown", this._onPointerDown.bind(this));
        document.addEventListener("keydown", this._onDocumentKeyDown.bind(this));
        document.addEventListener("keyup", this._onDocumentKeyUp.bind(this));
    }

    _onPointerMove(event) {
        this._pointer.set((event.clientX/window.innerWidth)*2-1, -(event.clientY/window.innerHeight)*2+1);
        this._raycaster.setFromCamera(this._pointer, this._camera);
        const intersects = this._raycaster.intersectObjects(this._objects);

        if(intersects.length > 0) {
            const intersect = intersects[0];
            this._rollOverMesh.position.copy(intersect.point).add(intersect.face.normal); // copy & add 유무에 대한 차이?
            this._rollOverMesh.position.divideScalar(50).floor().multiplyScalar(50).addScalar(25); // ?
        }
    }

    _onPointerDown(event) {
        this._pointer.set((event.clientX/window.innerWidth)*2-1, -(event.clientY/window.innerHeight)*2+1);
        this._raycaster.setFromCamera(this._pointer, this._camera);
        const intersects = this._raycaster.intersectObjects(this._objects);

        if(intersects.length > 0) {
            const intersect = intersects[0];

            if(this._isShiftDown) {
                if(intersect.object !== this._plane) {
                    this._scene.remove(intersect.object);
                    this._objects.splice(this._objects.indexOf(intersect.object), 1);
                }
            } else {
                const voxel = new THREE.Mesh(this._cubeGeo, this._cubeMaterial);
                console.log(intersect.point);
                voxel.position.copy(intersect.point).add(intersect.face.normal);
                voxel.position.divideScalar(50).floor().multiplyScalar(50).addScalar(25);
                this._scene.add(voxel);
                this._objects.push(voxel);
            }
        }
    }

    _onDocumentKeyDown(event) {
        switch(event.keyCode) {
            case 16: this._isShiftDown = true; break;
        }
    }

    _onDocumentKeyUp(event) {
        switch(event.keyCode) {
            case 16: this._isShiftDown = false; break;
        }
    }

    _setupPostprocessing() {
        //.
    }

    _setupModel() {
        const rollOverGeo = new THREE.BoxGeometry(50,50,50);
        const rollOverMaterial = new THREE.MeshBasicMaterial({ color: 0xff0000, opacity: 0.5, transparent: true });
        const rollOverMesh = new THREE.Mesh(rollOverGeo, rollOverMaterial);
        this._scene.add(rollOverMesh);

        const cubeGeo = new THREE.BoxGeometry(50, 50, 50);
        const cubeMaterial = new THREE.MeshLambertMaterial({color: 0xfeb74c, map: new THREE.TextureLoader().load("textures/square-outline-textured.png")});

        const gridHelper = new THREE.GridHelper(1000, 20);
        this._scene.add(gridHelper);

        const geometry = new THREE.PlaneGeometry(1000, 1000);
        geometry.rotateX(-Math.PI/2);
        const plane = new THREE.Mesh(geometry, new THREE.MeshBasicMaterial({visible: false}));
        this._scene.add(plane);

        this._objects.push(plane);

        this._rollOverMesh = rollOverMesh;
        this._plane = plane;
        this._cubeGeo = cubeGeo;
        this._cubeMaterial = cubeMaterial;
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(45, window.innerWidth / window.innerHeight, 1, 10000);
        camera.position.set(500, 800, 1300);
        camera.lookAt(0, 0, 0);

        this._camera = camera;
    }

    _setupControls() {
        //.
    }

    _setupLight() {
        const ambientLight = new THREE.AmbientLight(0x606060);
        this._scene.add(ambientLight);

        const directionalLight = new THREE.DirectionalLight(0xffffff);
        directionalLight.position.set(1, 0.75, 0.5).normalize();
        this._scene.add(directionalLight);

    }

    update() {
        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    new App();
}